@@ -8,11 +8,14 @@ import queryset
 from queryset import *
 import signals
 from signals import *
+from errors import *
+import errors
+import django
 
-__all__ = (document.__all__ + fields.__all__ + connection.__all__ +
-           queryset.__all__ + signals.__all__)
+__all__ = (list(document.__all__) + fields.__all__ + connection.__all__ +
+           list(queryset.__all__) + signals.__all__ + list(errors.__all__))
 
-VERSION = (0, 7, 10)
+VERSION = (0, 8, 0)
 
 
 def get_version():
